/*
		    TTTTTTT    OOOOO    7777777
		       T      O     O        7
		       T      O     O       7
		       T      O     O      7
		       T      O     O     7
		       T       OOOOO     7

			     EMULATEUR

			  Par Sylvain HUET
		(sylvain.huet@polytechnique.org)
				  1994-2002

  sound.c : gestion son pour windows

*/

#include <windows.h>
#include <mmsystem.h>

#include <memory.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "emuto7.h"
#include "monitor.h"

#define FREQ 44100

extern HWND TOhwnd;

// Global variables & type defs
typedef struct WaveOutBuffer
{
  WORD Rate;
  BYTE nChannels;
  LONG dwDataSize;
  BYTE Resolution;
  HWND hwnd;
  HGLOBAL hWaveHdr;
  LPWAVEHDR lpWaveHdr;
  HANDLE hData;  // handle of waveform data memory 
  HPSTR  lpData;  // pointer to waveform data memory 
} WaveOutBuffer;


HWAVEOUT hWaveOut;

#define NBUFFERS (3)
#define MAXBUFFERS (4)
int nPlayBuffers=NBUFFERS;
WaveOutBuffer WaveOutBuf[MAXBUFFERS];
int SendOutBuffers=0;
int PlayPending=-1;

BOOL StopPlayback=FALSE;
BOOL PlaybackInitialised=FALSE;

char bufsound[FREQ/EFREQ];
int isound=0;
int valsound=0;

void ClrSoundBuffer()
{
	int i;
	for(i=0;i<FREQ/EFREQ;i++) bufsound[i]=(char)128;
	isound=0;
	valsound=0;
}

int SoundOutFillBuffer(int OutBufToFill)
{
  char * S;
  int slen;

  S=bufsound;
  slen=FREQ/EFREQ;

// rcuprer l'chantillon  jouer dans S et slen
// par dfaut :
//  if (WaveOutBuf[OutBufToFill].Resolution==8)
//    memset(WaveOutBuf[OutBufToFill].lpData,128,WaveOutBuf[OutBufToFill].dwDataSize);
//  else
//    memset(WaveOutBuf[OutBufToFill].lpData,0,WaveOutBuf[OutBufToFill].dwDataSize);

  
  if (WaveOutBuf[OutBufToFill].lpData==NULL) return 0;

  if (slen>WaveOutBuf[OutBufToFill].dwDataSize)
  {
      slen=WaveOutBuf[OutBufToFill].dwDataSize;
  }
  else if (WaveOutBuf[OutBufToFill].Resolution==8)
	  memset(WaveOutBuf[OutBufToFill].lpData,128,WaveOutBuf[OutBufToFill].dwDataSize);
  else
	  memset(WaveOutBuf[OutBufToFill].lpData,0,WaveOutBuf[OutBufToFill].dwDataSize);
  memcpy(WaveOutBuf[OutBufToFill].lpData,S,slen);  
  return 0;
}


void WaveOutFreeAll()
{
  int i;
  if (!PlaybackInitialised) return;
  for (i=0;i<nPlayBuffers;i++)
  {
    waveOutUnprepareHeader(hWaveOut,WaveOutBuf[i].lpWaveHdr,sizeof(WAVEHDR));

    if (WaveOutBuf[i].lpData != NULL)
        GlobalUnlock(WaveOutBuf[i].hData);
    if (WaveOutBuf[i].hData) 
      GlobalFree(WaveOutBuf[i].hData);

    if (WaveOutBuf[i].lpWaveHdr != NULL)
      GlobalUnlock(WaveOutBuf[i].hWaveHdr);
    if (WaveOutBuf[i].hWaveHdr)
      GlobalFree(WaveOutBuf[i].hWaveHdr);
  }
  memset(WaveOutBuf,0,nPlayBuffers*sizeof(WaveOutBuffer));
  PlaybackInitialised=FALSE;
  SendOutBuffers=0;
}


int SoundInitPlayBack(int rate,int nChannels,int ChannelSize,int BitsPerSample)
{
  UINT wResult;
  PCMWAVEFORMAT pcmWaveFormat;
  int i;

  if (PlaybackInitialised) return 0;

  // Open a waveform device for output using window callback.

  if (BitsPerSample<8) BitsPerSample=8;
  if (BitsPerSample>8) BitsPerSample=16;
  pcmWaveFormat.wf.wFormatTag=WAVE_FORMAT_PCM;
  pcmWaveFormat.wf.nChannels=nChannels;
  pcmWaveFormat.wf.nSamplesPerSec=rate;
  pcmWaveFormat.wf.nAvgBytesPerSec=rate*nChannels*(BitsPerSample/8);
  pcmWaveFormat.wf.nBlockAlign=(nChannels*BitsPerSample)/8;
  pcmWaveFormat.wBitsPerSample=BitsPerSample;

  wResult=waveOutOpen((LPHWAVEOUT)&hWaveOut, WAVE_MAPPER,
                  (LPWAVEFORMATEX)&pcmWaveFormat,
                  (LONG)TOhwnd, 0L, CALLBACK_WINDOW);
  if (wResult != 0)
  {
	return wResult;
  }

  ClrSoundBuffer();
  for (i=0;i<nPlayBuffers;i++)
  {
    WaveOutBuf[i].nChannels=nChannels;
    WaveOutBuf[i].Rate=rate;
    WaveOutBuf[i].Resolution=BitsPerSample;
    WaveOutBuf[i].hwnd=TOhwnd;
    WaveOutBuf[i].dwDataSize=nChannels*(BitsPerSample/8)*ChannelSize;
    
    // Allocate and lock memory for the waveform data.
    WaveOutBuf[i].hData = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, WaveOutBuf[i].dwDataSize );
    if (!WaveOutBuf[i].hData)
    {
//        MMechostr(MSKDEBUG,"PlayBack: Unable to Allocate memory\n");
        return -1;
    }
    if ((WaveOutBuf[i].lpData = GlobalLock(WaveOutBuf[i].hData)) == NULL)
	{
//        MMechostr(MSKDEBUG,"PlayBack: Unable to Lock Memory\n");
        WaveOutFreeAll();
        return -1;
    }

    // Allocate and lock memory for the header.
    WaveOutBuf[i].hWaveHdr = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE,(DWORD) sizeof(WAVEHDR));
    if (WaveOutBuf[i].hWaveHdr == NULL)
    {
        WaveOutFreeAll();
//        MMechostr(MSKDEBUG,"PlayBack: Unable to allocate wave header\n");
        return -1;
    }

    WaveOutBuf[i].lpWaveHdr = (LPWAVEHDR) GlobalLock(WaveOutBuf[i].hWaveHdr);
    if (WaveOutBuf[i].lpWaveHdr == NULL)
    {
        WaveOutFreeAll();
//        MMechostr(MSKDEBUG,"PlayBack: Unable to Lock wave header memory\n");
        return -1;
    }

    // After allocation, set up and prepare header.

    WaveOutBuf[i].lpWaveHdr->lpData = WaveOutBuf[i].lpData;
    WaveOutBuf[i].lpWaveHdr->dwBufferLength = WaveOutBuf[i].dwDataSize;
    WaveOutBuf[i].lpWaveHdr->dwUser = i;
    WaveOutBuf[i].lpWaveHdr->dwFlags = 0L;
    WaveOutBuf[i].lpWaveHdr->dwLoops = 1L;
    wResult = waveOutPrepareHeader(hWaveOut,WaveOutBuf[i].lpWaveHdr,sizeof(WAVEHDR));
    if (wResult != 0)
    {
      waveOutClose(hWaveOut);
      WaveOutFreeAll();
      return 0;
    }
//	MMechostr(MSKDEBUG,"PrepareOK\n");

    SoundOutFillBuffer(i);
  }

  for (i=0;i<nPlayBuffers;i++)
  {
    // Now the data block can be sent to the output device. The
    // waveOutWrite function returns immediately and waveform
    // data is sent to the output device in the background.
    wResult = waveOutWrite(hWaveOut,WaveOutBuf[i].lpWaveHdr,sizeof(WAVEHDR));
    if (wResult != 0)
    {
      wResult=waveOutReset(hWaveOut);
      wResult=waveOutClose(hWaveOut);
      WaveOutFreeAll();
      return -1;
    }
    SendOutBuffers++;
  }
  StopPlayback=FALSE;
  PlaybackInitialised=TRUE;
  return 1;
}

int SoundNextPlayBack(int i)
{
  WORD wResult;
 
  SendOutBuffers--;
  if (!StopPlayback)
  {
    SoundOutFillBuffer(i);
    wResult = waveOutWrite(hWaveOut,WaveOutBuf[i].lpWaveHdr,sizeof(WAVEHDR));
    if (wResult != 0)
    {
      WaveOutFreeAll();
//      MMechostr(MSKDEBUG,"\nPlayBackCB: Unable to write block to the device\n");
      return 0;
    }
    SendOutBuffers++;
  }
  else
  {
    if (!SendOutBuffers)
    {
      WaveOutFreeAll();
      StopPlayback=FALSE;
      waveOutClose(hWaveOut);
	}
  }
  return 1;
}


/* API appelable de l'extrieur */
int SoundPlayEvent(int hwnd,int msg,int wParam,int lParam)
{
  return SoundNextPlayBack(((LPWAVEHDR)lParam)->dwUser);
}

void BeginSoundBuffer()
{
	isound=0;
	countop=0;
}

void FinishSoundBuffer()
{
	int i;
	char *p;

	p=bufsound;
	for(i=isound;i<FREQ/EFREQ;i++) p[i]=valsound;
}

void NewSoundBuffer()
{
	int i;
	char *p;

	p=bufsound;
	for(i=isound;i<FREQ/EFREQ;i++) p[i]=valsound;
	isound=0;
}

void SetSound(int val)
{
	int newisound;
	int i;
	char *p;

	if (!PlaybackInitialised) return;
	p=bufsound;
	newisound=countop*FREQ/EFREQ/(330000/EFREQ);
	if (newisound>FREQ/EFREQ) newisound=FREQ/EFREQ;
	for(i=isound;i<newisound;i++) p[i]=valsound;
	if (val) valsound=136;
	else valsound=120;
	isound=newisound;
}

void StartSound()
{
	SoundInitPlayBack(FREQ,1,FREQ/EFREQ,8);
}

void StopSound()
{
	WaveOutFreeAll();
}

int ifSoundOn()
{
	if (PlaybackInitialised) return 1;
	return 0;
}
